/***************************************************************************
 *   copyright           : (C) 2004 by Hendrik Sattler                     *
 *   mail                : post@hendrik-sattler.de                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ATCOMMAND_H
#define ATCOMMAND_H

#include "stdlib.h"

/* MUST be run before any other at_* function is used
 */
void at_init (char* pin);

/* useful wrapper around tty_read_line() that removes
 * echos, empty lines and does pin handling
 */
char* at_read_line ();

/* used as callback for tty_read_line()
 */
int at_check_line_end (const char* input, const size_t len);

/* May be used directly for simple commands of the form
 *   AT<command>[=<parmlist>]
 * command must not be NULL
 * parmlist may be NULL
 */
void at_command_send (const char* command, const char* parmlist);

/* May be used by syntax/value requests of the form
 *   AT<command>?
 * command must not be NULL
 */
void at_syntax_request (const char* command);

/* used by commands that expect only two lines in the response
 * - first line for the data
 * - second line for command status
 * This returns the  data or NULL on any error.
 * A possible command prefix is removed from the data.
 * command may be NULL (avoids prefix removal)
 */
char* at_get_value (const char* command);

char** at_parse_stringlist (char* stringlist);


/* ftype must not be NULL (checked)
 * if part or total is 0 (means deleting) the value of total is ignored
 * if part or total is negative, part and total are ignored
 */
void at_sie_binary_write (const char* ftype, unsigned int slot,
			  int part, int total);

/* ftype must not be NULL (checked)
 */
void at_sie_binary_read (const char* ftype, unsigned int slot);

/* if number or text are NULL, both are ignored
 */
void at_gen_phonebook_write (unsigned int slot,
			     const char* number,
			     unsigned int numtype,
			     const char* text);

void at_gen_phonebook_read (unsigned int slot_low,
			    unsigned int slot_high);
void at_sie_phonebook_read_sorted (unsigned int slot_low,
				   unsigned int slot_high);
void at_sie_phonebook_vcf_read (unsigned int slot, int type);

/* mem must not be NULL (checked)
 */
void at_gen_phonebook_select (const char* mem);
void at_sie_phonebook_select (const char* mem);

void at_gen_sms_delete (unsigned int slot);

/* if number is NULL, it is ignored
 */
void at_gen_sms_slot_send (unsigned int slot,
			   const char* number);

void at_gen_sms_send (unsigned int len);
void at_gen_sms_store (unsigned int len);

/* mem1 must not be NULL (checked
 * if mem2 is NULL, mem2 and mem3 is ignored
 * if mem3 is NULL, mem is ignored
 */
void at_gen_sms_mem (const char* mem1,
		     const char* mem2,
		     const char* mem3);


void at_gen_sms_slot_read (unsigned int slot);
void at_sie_sms_slot_read (unsigned int slot);
void at_gen_sms_list_read (unsigned int list_id);
void at_sie_sms_list_read (unsigned int list_id);


void at_gen_charset (const char* charset);
void at_gen_smsc (const char* number);
void at_gen_pin (const char* pin);


//common prefixes
#define AT_PREFIX         "AT"
#define AT_GENERIC_PREFIX "+C"
#define AT_SIEMENS_PREFIX "^S"

//common delimiters
#define AT_WRITE_DELIM  "="
//#define AT_WRITE(at_command)         AT_PREFIX""at_command""AT_WRITE_DELIM

//common suffixes
#define AT_READ_SUFFIX   "?"
#define AT_REPLY_SUFFIX  ": "
//#define AT_READ(at_command)          AT_PREFIX""at_command""AT_READ_SUFFIX
//#define AT_TEST(at_command)          AT_WRITE(at_command)""AT_READ_SUFFIX
#define AT_REPLY_PREFIX(at_command)  at_command""AT_REPLY_SUFFIX


/* often, commands are prefixed to a reply
 * or the following status messages appear
 */
#define AT_OK         "OK"
#define AT_ERROR      "ERROR"
#define AT_ERROR_CME  AT_REPLY_PREFIX("+CME ERROR")
#define AT_ERROR_CMS  AT_REPLY_PREFIX("+CMS ERROR")

enum return_code {
  AT_RET_ERROR = 0x01, //also mask for errors
  AT_RET_ERROR_CME = 0x03,
  AT_RET_ERROR_CMS = 0x05,
  AT_RET_OK = 0x00,
  AT_RET_ANSWER = 0x02, //answer line for a command
  AT_RET_OTHER = 0x04//none of the above
};

/* command may be NULL
 * if a prefix was detected and content is not NULL
 *   *content is a pointer to the first character after
 *   the prefix in response
 * if no prefix was detectet, *content points to response
 */
enum return_code at_line_type (const char* response,
			       const char* command,
			       char** content);

/* generic commands
 * those may work for phones from all vendors
 */
//those have no parameters at all
#define AT_GEN_INIT        "Z"     // init the phone to defaults
#define AT_GEN_VENDOR      "+CGMI" // get vendor
#define AT_GEN_MODEL       "+CGMM" // get model
#define AT_GEN_REVISION    "+CGMR" // get model revision
#define AT_GEN_IMSI        "+CGSN" // get international mobile subcriber identity ?????
#define AT_GEN_IMEI        "+CIMI" // get international mobile entity identity
#define AT_GEN_BATTERY     "+CBC"  // get battery status
#define AT_GEN_SIGNAL      "+CSQ"  // get signal strength/ber

//these ones may use the parameter scheme ?/=?/=<...>
#define AT_GEN_ECHO_ENABLE   "E1"    // enable echo response
#define AT_GEN_ECHO_DISABLE  "E0"    // disable echo response
#define AT_GEN_ERROR         "+CMEE" // select type of response for errors
#define AT_GEN_PB_SELECT     "+CPBS" // select phonebook memory / get list of phonebook memories
#define AT_GEN_PB_READ       "+CPBR" // read entries from phonebook
#define AT_GEN_PB_WRITE      "+CPBW" // write entries to phonebook
#define AT_GEN_SMS_MEM       "+CPMS" // select SMS memory / get list
#define AT_GEN_SMS_STORE     "+CMGW" // store SMS to slot
#define AT_GEN_SMS_DELETE    "+CMGD" // delete SMS slot
#define AT_GEN_SMS_SEND      "+CMGS" // send SMS (without storing to a slot)
#define AT_GEN_SMS_SLOT_SEND "+CMSS" // send SMS from slot
#define AT_GEN_SMS_SLOT_READ "+CMGR" // read SMS from slot (marks unread as read)
#define AT_GEN_SMS_LIST_READ "+CMGL" // read SMS by status (marks unread as read)
#define AT_GEN_SMS_PHASE     "+CSMS"
#define AT_GEN_MSG_DIRECT    "+CNMI"
#define AT_GEN_MSG_ACK       "+CNMA"
#define AT_GEN_SMSC          "+CSCA" // set/get SMSC
#define AT_GEN_CHARSET       "+CSCS" // set/get current character set
#define AT_GEN_PIN           "+CPIN" // enter pin (DANGEROUS)
#define AT_GEN_TIME          "+CCLK" // set/read time
#define AT_GEN_OPERATOR      "+COPS" // get operator
#define AT_GEN_CSD_STATUS    "+CREG" // get area code and cell id (for CSD)
#define AT_GEN_GPRS_STATUS   "+CGREG" // get area code and cell id (for GPRS)
#define AT_GEN_GPRS_ATTACH   "+CGATT" // is GPRS in attached mode?
#define AT_GEN_GPRS_CLASS    "+CGCLASS" //get GPRS class

/* siemens-specific commands
 * probably only works with phones made by Siemens
 * not all of those are (really) fully documented
 */
#define AT_SIE_PB_VCF_READ    "^SDBR"
#define AT_SIE_PB_READ_SORTED "^SPBG"
#define AT_SIE_PB_SELECT      "^SPBS" // select phonebook memory / get list of phonebook memories
#define AT_SIE_PB_DELETE      "^SDLD" // delete for phonebooks (only DC, MD and LD)
#define AT_SIE_CARD_ID        "^SCID" // get SIM card identity number
#define AT_SIE_BIN_READ       "^SBNR" // read binary data 
#define AT_SIE_BIN_WRITE      "^SBNW" // write binary data
#define AT_SIE_SMS_SLOT_READ  "^SMGR" // read SMS from slot (without marking unread as read)
#define AT_SIE_SMS_LIST_READ  "^SMGL" // read SMS by list (without marking unread as read)

#endif
